from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast

import daft
import daft.exceptions
import daft.functions as F
from daft import Expression
from narwhals._utils import (
    Implementation,
    ValidateBackendVersion,
    Version,
    check_column_names_are_unique,
    extend_bool,
    generate_temporary_column_name,
    not_implemented,
    parse_columns_to_drop,
)
from narwhals.exceptions import (
    ColumnNotFoundError,
    DuplicateError,
    MultiOutputExpressionError,
)
from narwhals.typing import CompliantLazyFrame

from narwhals_daft.group_by import DaftLazyGroupBy
from narwhals_daft.utils import evaluate_exprs, lit, native_to_narwhals_dtype

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence
    from io import BytesIO
    from pathlib import Path
    from types import ModuleType

    from narwhals._compliant.typing import CompliantDataFrameAny
    from narwhals._utils import _LimitedContext
    from narwhals.dataframe import LazyFrame
    from narwhals.dtypes import DType
    from narwhals.typing import JoinStrategy
    from typing_extensions import TypeIs

    from narwhals_daft.expr import DaftExpr, WindowInputs
    from narwhals_daft.namespace import DaftNamespace


class DaftLazyFrame(
    CompliantLazyFrame["DaftExpr", "daft.DataFrame", "LazyFrame[daft.DataFrame]"],
    ValidateBackendVersion,
):
    _implementation = Implementation.UNKNOWN

    def __init__(
        self,
        native_dataframe: daft.DataFrame,
        *,
        version: Version,
        validate_backend_version: bool = False,
    ) -> None:
        self._native_frame: daft.DataFrame = native_dataframe
        self._version = version
        self._cached_schema: dict[str, DType] | None = None
        self._cached_columns: list[str] | None = None
        if validate_backend_version:  # pragma: no cover
            self._validate_backend_version()

    @staticmethod
    def _is_native(obj: daft.DataFrame | Any) -> TypeIs[daft.DataFrame]:
        return isinstance(obj, daft.DataFrame)

    @classmethod
    def from_native(
        cls, data: daft.DataFrame, /, *, context: _LimitedContext
    ) -> DaftLazyFrame:
        return cls(data, version=context._version)

    def to_narwhals(self) -> LazyFrame[daft.DataFrame]:
        return self._version.lazyframe(self, level="lazy")

    def __native_namespace__(self) -> ModuleType:
        import daft

        return daft

    def __narwhals_namespace__(self) -> DaftNamespace:
        from narwhals_daft.namespace import DaftNamespace

        return DaftNamespace(version=self._version)

    def __narwhals_lazyframe__(self) -> DaftLazyFrame:
        return self

    def _with_version(self, version: Version) -> DaftLazyFrame:
        return self.__class__(self._native_frame, version=version)

    def _with_native(self, df: daft.DataFrame) -> DaftLazyFrame:
        return self.__class__(df, version=self._version)

    def _iter_columns(self) -> Iterator[daft.Expression]:
        return iter(self._native_frame.columns)

    def _evaluate_expr(self, expr: DaftExpr) -> Expression:
        result: Sequence[Expression] = expr(self)
        assert len(result) == 1  # debug assertion  # noqa: S101
        return result[0]

    def _evaluate_window_expr(
        self, expr: DaftExpr, /, window_inputs: WindowInputs
    ) -> Expression:
        result = expr.window_function(self, window_inputs)
        if len(result) != 1:  # pragma: no cover
            msg = "multi-output expressions not allowed in this context"
            raise MultiOutputExpressionError(msg)
        return result[0]

    @property
    def columns(self) -> list[str]:
        if self._cached_columns is None:
            self._cached_columns = (
                list(self.schema)
                if self._cached_schema is not None
                else self.native.column_names
            )
        return self._cached_columns

    def collect(
        self, backend: ModuleType | Implementation | str | None, **kwargs: Any
    ) -> CompliantDataFrameAny:
        if backend is None or backend is Implementation.PYARROW:
            from narwhals._arrow.dataframe import ArrowDataFrame

            return ArrowDataFrame(
                native_dataframe=self._native_frame.to_arrow(),
                validate_backend_version=True,
                version=self._version,
                validate_column_names=True,
            )

        if backend is Implementation.PANDAS:
            from narwhals._pandas_like.dataframe import PandasLikeDataFrame

            return PandasLikeDataFrame(
                native_dataframe=self._native_frame.to_pandas(),
                implementation=Implementation.PANDAS,
                validate_backend_version=True,
                version=self._version,
                validate_column_names=True,
            )

        if backend is Implementation.POLARS:
            import polars as pl  # ignore-banned-import
            from narwhals._polars.dataframe import PolarsDataFrame

            return PolarsDataFrame(
                df=cast("pl.DataFrame", pl.from_arrow(self._native_frame.to_arrow())),
                validate_backend_version=True,
                version=self._version,
            )

        msg = f"Unsupported `backend` value: {backend}"  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover

    def simple_select(self, *column_names: str) -> DaftLazyFrame:
        return self._with_native(self._native_frame.select(*column_names))

    def aggregate(self, *exprs: DaftExpr) -> DaftLazyFrame:
        new_columns_map = evaluate_exprs(self, *exprs)
        return self._with_native(
            self._native_frame.agg([val.alias(col) for col, val in new_columns_map])
        )

    def select(self, *exprs: DaftExpr) -> DaftLazyFrame:
        new_columns_map = evaluate_exprs(self, *exprs)
        if not new_columns_map:
            msg = "At least one expression must be passed to LazyFrame.select"
            raise ValueError(msg)
        try:
            return self._with_native(
                self._native_frame.select(
                    *(val.alias(col) for col, val in new_columns_map)
                )
            )
        except daft.exceptions.DaftCoreException as e:
            if "duplicate" in str(e):  # pragma: no cover
                raise DuplicateError(e) from None
            if "not found" in str(e):
                msg = (
                    f"{e!s}\n\nHint: Did you mean one of these columns: {self.columns}?"
                )
                raise ColumnNotFoundError(msg) from e
            raise

    def with_columns(self, *exprs: DaftExpr) -> DaftLazyFrame:
        new_columns_map = dict(evaluate_exprs(self, *exprs))
        return self._with_native(self._native_frame.with_columns(new_columns_map))

    def filter(self, predicate: DaftExpr) -> DaftLazyFrame:
        # `[0]` is safe as the predicate's expression only returns a single column
        mask = predicate._call(self)[0]
        return self._with_native(self._native_frame.filter(mask))

    @property
    def schema(self) -> dict[str, DType]:
        if self._cached_schema is None:
            # Note: prefer `self._cached_schema` over `functools.cached_property`
            # due to Python3.13 failures.
            self._cached_schema = {
                field.name: native_to_narwhals_dtype(field.dtype, self._version)
                for field in (self._native_frame.schema())
            }
        return self._cached_schema

    def collect_schema(self) -> dict[str, DType]:
        return {
            field.name: native_to_narwhals_dtype(field.dtype, self._version)
            for field in self._native_frame.schema()
        }

    def drop(self, columns: Sequence[str], *, strict: bool) -> DaftLazyFrame:
        columns_to_drop = parse_columns_to_drop(self, columns, strict=strict)
        selection = [col for col in self.columns if col not in columns_to_drop]
        return self._with_native(self._native_frame.select(*selection))

    def head(self, n: int) -> DaftLazyFrame:
        return self._with_native(self._native_frame.limit(n))

    def sort(
        self, *by: str, descending: bool | Sequence[bool], nulls_last: bool
    ) -> DaftLazyFrame:
        return self._with_native(
            self._native_frame.sort(
                list(by),
                desc=descending if isinstance(descending, bool) else list(descending),
                nulls_first=not nulls_last,
            )
        )

    def top_k(
        self, k: int, *, by: Iterable[str], reverse: bool | Sequence[bool]
    ) -> DaftLazyFrame:
        descending = [not x for x in extend_bool(reverse, len(list(by)))]
        return self._with_native(
            self._native_frame.sort(list(by), desc=descending, nulls_first=False).limit(
                k
            )
        )

    def drop_nulls(self, subset: Sequence[str] | None) -> DaftLazyFrame:
        if subset:
            return self._with_native(self._native_frame.drop_null(*subset))
        return self._with_native(self._native_frame.drop_null())

    def rename(self, mapping: Mapping[str, str]) -> DaftLazyFrame:
        selection = [
            daft.col(col).alias(mapping[col]) if col in mapping else col
            for col in self.columns
        ]
        return self._with_native(self.native.select(*selection))

    def unique(
        self, subset: Sequence[str] | None, *, keep: str, order_by: Sequence[str] | None
    ) -> DaftLazyFrame:
        subset_ = subset or self.columns
        tmp_name = generate_temporary_column_name(8, self.columns, prefix="row_index_")
        window = daft.Window().partition_by(*subset_)
        if order_by and keep == "last":
            window = window.order_by(*order_by, desc=True, nulls_first=False)
        elif order_by:
            window = window.order_by(*order_by, desc=False, nulls_first=True)
        if keep == "none":
            expr = daft.col(self.columns[0]).count(mode="all").over(window)
        else:
            expr = F.row_number().over(window)
        df = self.native.with_column(tmp_name, expr).filter(
            daft.col(tmp_name) == lit(1)
        )
        df = df.select(*(x for x in df.column_names if x != tmp_name))
        return self._with_native(df)

    def join(
        self,
        other: DaftLazyFrame,
        how: JoinStrategy,
        left_on: Sequence[str] | None,
        right_on: Sequence[str] | None,
        suffix: str,
    ) -> DaftLazyFrame:
        if how == "cross":
            return self._with_native(
                self.native.join(other.native, how="cross", prefix="", suffix=suffix)
            )
        left_columns = self.columns
        right_columns = other.columns

        right_on_ = list(right_on) if right_on is not None else []
        left_on_ = list(left_on) if left_on is not None else []

        # create a mapping for columns on other
        # `right_on` columns will be renamed as `left_on`
        # the remaining columns will be either added the suffix or left unchanged.
        right_cols_to_rename = (
            [c for c in right_columns if c not in right_on_]
            if how != "full"
            else right_columns
        )

        rename_mapping = {
            **dict(zip(right_on_, left_on_, strict=True)),
            **{
                colname: f"{colname}{suffix}" if colname in left_columns else colname
                for colname in right_cols_to_rename
            },
        }
        plx = self.__narwhals_namespace__()
        other_native = other.select(
            *[plx.col(old).alias(new) for old, new in rename_mapping.items()]
        ).native
        col_order = left_columns.copy()

        if how in {"inner", "left", "cross"}:
            col_order.extend(
                rename_mapping[colname]
                for colname in right_columns
                if colname not in right_on_
            )
        elif how == "full":
            col_order.extend(rename_mapping.values())

        check_column_names_are_unique(col_order)

        right_on_remapped = [rename_mapping[c] for c in right_on_]
        how_native: Literal["inner", "left", "semi", "anti", "outer"] = (
            "outer" if how == "full" else how
        )

        return self._with_native(
            self.native.join(
                other_native,
                left_on=[daft.col(x) for x in left_on_],
                right_on=[daft.col(x) for x in right_on_remapped],
                how=how_native,
            ).select(*col_order)
        )

    def unpivot(
        self,
        on: Sequence[str] | None,
        index: Sequence[str] | None,
        variable_name: str,
        value_name: str,
    ) -> DaftLazyFrame:
        index_ = [] if index is None else index
        on_ = [c for c in self.columns if c not in index_] if on is None else on
        return self._with_native(
            self._native_frame.unpivot(
                ids=index_,
                values=on_,
                variable_name=variable_name,
                value_name=value_name,
            )
        )

    def with_row_index(self, name: str, order_by: Sequence[str]) -> DaftLazyFrame:
        row_index_expr = (
            (
                daft.functions.row_number().over(
                    daft.Window().partition_by(lit(1)).order_by(*order_by)
                )
                - 1
            )
            if order_by
            else daft.functions.monotonically_increasing_id()
        )
        return self._with_native(
            self.native.select(row_index_expr.alias(name), *self.columns)
        )

    def group_by(
        self, keys: Sequence[str] | Sequence[DaftExpr], *, drop_null_keys: bool
    ) -> DaftLazyGroupBy:
        return DaftLazyGroupBy(self, keys, drop_null_keys=drop_null_keys)

    def explode(self, columns: Sequence[str]) -> DaftLazyFrame:
        return self._with_native(self.native.explode(*columns))

    def sink_parquet(self, file: str | Path | BytesIO) -> None:
        self.native.write_parquet(str(file))

    gather_every = not_implemented.deprecated(
        "`LazyFrame.gather_every` is deprecated and will be removed in a future version."
    )
    join_asof = not_implemented()
    tail = not_implemented.deprecated(
        "`LazyFrame.tail` is deprecated and will be removed in a future version."
    )



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from narwhals._compliant.group_by import CompliantGroupBy
from narwhals._utils import is_sequence_of

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence

    from daft import Expression

    from narwhals_daft.dataframe import DaftLazyFrame
    from narwhals_daft.expr import DaftExpr


def _evaluate_aliases_single(frame: DaftLazyFrame, expr: DaftExpr) -> Sequence[str]:
    # NOTE: Ignore intermittent [False Negative]
    # Argument of type "CompliantFrameT@ImplExpr" cannot be assigned to parameter of type "CompliantFrameT@ImplExpr"
    #  Type "CompliantFrameT@ImplExpr" is not assignable to type "CompliantFrameT@ImplExpr"
    names = expr._evaluate_output_names(frame)  # pyright: ignore[reportArgumentType]
    return alias(names) if (alias := expr._alias_output_names) else names


def _evaluate_aliases_many(
    frame: DaftLazyFrame, exprs: Iterable[DaftExpr], /
) -> list[str]:
    it = (_evaluate_aliases_single(frame, expr) for expr in exprs)
    return list(chain.from_iterable(it))


class ParseKeysGroupBy(CompliantGroupBy["DaftLazyFrame", "DaftExpr"]):
    def _parse_keys(
        self, compliant_frame: DaftLazyFrame, keys: Sequence[DaftExpr] | Sequence[str]
    ) -> tuple[DaftLazyFrame, list[str], list[str]]:
        if is_sequence_of(keys, str):
            keys_str = list(keys)
            return compliant_frame, keys_str, keys_str.copy()
        return self._parse_expr_keys(compliant_frame, keys=keys)

    @staticmethod
    def _parse_expr_keys(
        compliant_frame: DaftLazyFrame, keys: Sequence[DaftExpr]
    ) -> tuple[DaftLazyFrame, list[str], list[str]]:
        """Parses key expressions to set up `.agg` operation with correct information.

        Since keys are expressions, it's possible to alias any such key to match
        other dataframe column names.

        In order to match polars behavior and not overwrite columns when evaluating keys:

        - We evaluate what the output key names should be, in order to remap temporary column
            names to the expected ones, and to exclude those from unnamed expressions in
            `.agg(...)` context (see https://github.com/narwhals-dev/narwhals/pull/2325#issuecomment-2800004520)
        - Create temporary names for evaluated key expressions that are guaranteed to have
            no overlap with any existing column name.
        - Add these temporary columns to the compliant dataframe.
        """
        tmp_name_length = max(len(str(c)) for c in compliant_frame.columns) + 1

        def _temporary_name(key: str) -> str:
            # 5 is the length of `__tmp`
            key_str = str(key)  # pandas allows non-string column names :sob:
            return f"_{key_str}_tmp{'_' * (tmp_name_length - len(key_str) - 5)}"

        keys_aliases = [
            _evaluate_aliases_single(compliant_frame, expr) for expr in keys
        ]
        safe_keys = [
            # multi-output expression cannot have duplicate names, hence it's safe to suffix
            key.name.map(_temporary_name)
            if (metadata := key._metadata) and metadata.expansion_kind.is_multi_output()
            # otherwise it's single named and we can use Expr.alias
            else key.alias(_temporary_name(new_names[0]))
            for key, new_names in zip(keys, keys_aliases, strict=True)
        ]
        return (
            compliant_frame.with_columns(*safe_keys),
            _evaluate_aliases_many(compliant_frame, safe_keys),
            list(chain.from_iterable(keys_aliases)),
        )


class DaftLazyGroupBy(ParseKeysGroupBy, CompliantGroupBy["DaftLazyFrame", "DaftExpr"]):
    def __init__(
        self,
        df: DaftLazyFrame,
        keys: Sequence[DaftExpr] | Sequence[str],
        /,
        *,
        drop_null_keys: bool,
    ) -> None:
        frame, self._keys, self._output_key_names = self._parse_keys(df, keys=keys)
        self._compliant_frame = (
            frame.drop_nulls(self._keys) if drop_null_keys else frame
        )

    def _evaluate_expr(self, expr: DaftExpr, /) -> Iterator[Expression]:
        output_names = expr._evaluate_output_names(self.compliant)
        aliases = (
            expr._alias_output_names(output_names)
            if expr._alias_output_names
            else output_names
        )
        native_exprs = expr(self.compliant)
        if expr._is_multi_output_unnamed():
            exclude = {*self._keys, *self._output_key_names}
            for native_expr, name, alias in zip(
                native_exprs, output_names, aliases, strict=True
            ):
                if name not in exclude:
                    yield expr._alias_native(native_expr, alias)
        else:
            for native_expr, alias in zip(native_exprs, aliases, strict=True):
                yield expr._alias_native(native_expr, alias)

    def _evaluate_exprs(self, exprs: Iterable[DaftExpr], /) -> Iterator[Expression]:
        for expr in exprs:
            yield from self._evaluate_expr(expr)

    def agg(self, *exprs: DaftExpr) -> DaftLazyFrame:
        result = (
            self.compliant.native.groupby(*self._keys).agg(*agg_columns)
            if (agg_columns := tuple(self._evaluate_exprs(exprs)))
            else self.compliant.native.select(*self._keys).drop_duplicates()
        )

        return self.compliant._with_native(result).rename(
            dict(zip(self._keys, self._output_key_names, strict=True))
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import daft.functions as F
from narwhals._compliant.any_namespace import DateTimeNamespace
from narwhals._utils import not_implemented

if TYPE_CHECKING:
    from narwhals_daft.expr import DaftExpr


class ExprDateTimeNamesSpace(DateTimeNamespace["DaftExpr"]):
    def __init__(self, expr: DaftExpr, /) -> None:
        self._compliant = expr

    @property
    def compliant(self) -> DaftExpr:
        return self._compliant

    def date(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.date(expr))

    def year(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.year(expr))

    def month(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.month(expr))

    def day(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.day(expr))

    def hour(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.hour(expr))

    def minute(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.minute(expr))

    def second(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.second(expr))

    def millisecond(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.millisecond(expr))

    def microsecond(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.microsecond(expr))

    def nanosecond(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.nanosecond(expr))

    def weekday(self) -> DaftExpr:
        return self.compliant._with_elementwise(
            lambda expr: F.day_of_week(expr) + 1
        )  # daft is 0-6

    def ordinal_day(self) -> DaftExpr:
        return self.compliant._with_elementwise(lambda expr: F.day_of_year(expr))

    to_string = not_implemented()
    replace_time_zone = not_implemented()
    convert_time_zone = not_implemented()
    timestamp = not_implemented()
    total_minutes = not_implemented()
    total_seconds = not_implemented()
    total_milliseconds = not_implemented()
    total_microseconds = not_implemented()
    total_nanoseconds = not_implemented()
    truncate = not_implemented()
    offset_by = not_implemented()



from __future__ import annotations

from typing import TYPE_CHECKING

import daft
from daft import DataType
from narwhals._utils import isinstance_or_issubclass

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from narwhals._utils import Version
    from narwhals.dtypes import DType

    from narwhals_daft.dataframe import DaftLazyFrame
    from narwhals_daft.expr import DaftExpr

lit = daft.lit
"""Alias for `daft.lit`."""


def evaluate_exprs(
    df: DaftLazyFrame, /, *exprs: DaftExpr
) -> list[tuple[str, daft.Expression]]:
    native_results: list[tuple[str, daft.Expression]] = []
    for expr in exprs:
        native_series_list = expr._call(df)
        output_names = expr._evaluate_output_names(df)
        if expr._alias_output_names is not None:
            output_names = expr._alias_output_names(output_names)
        if len(output_names) != len(native_series_list):  # pragma: no cover
            msg = f"Internal error: got output names {output_names}, but only got {len(native_series_list)} results"
            raise AssertionError(msg)
        native_results.extend(zip(output_names, native_series_list, strict=True))
    return native_results


def native_to_narwhals_dtype(daft_dtype: DataType, version: Version) -> DType:  # noqa: PLR0912,C901,PLR0911
    dtypes = version.dtypes

    if daft_dtype == DataType.int64():
        return dtypes.Int64()
    if daft_dtype == DataType.int32():
        return dtypes.Int32()
    if daft_dtype == DataType.int16():
        return dtypes.Int16()
    if daft_dtype == DataType.int8():
        return dtypes.Int8()
    if daft_dtype == DataType.uint64():
        return dtypes.UInt64()
    if daft_dtype == DataType.uint32():
        return dtypes.UInt32()
    if daft_dtype == DataType.uint16():
        return dtypes.UInt16()
    if daft_dtype == DataType.uint8():
        return dtypes.UInt8()
    if daft_dtype == DataType.float64():
        return dtypes.Float64()
    if daft_dtype == DataType.float32():
        return dtypes.Float32()
    if daft_dtype == DataType.string():
        return dtypes.String()
    if daft_dtype == DataType.date():
        return dtypes.Date()
    if daft_dtype == DataType.timestamp("us", None):
        return dtypes.Datetime("us", None)
    if daft_dtype == DataType.bool():
        return dtypes.Boolean()
    if daft_dtype == DataType.duration("us"):
        return dtypes.Duration("us")
    if daft_dtype == DataType.decimal128(1, 1):  # pragma: no cover
        return dtypes.Decimal()
    if DataType.is_fixed_size_list(daft_dtype):
        return dtypes.Array(
            native_to_narwhals_dtype(daft_dtype.dtype, version), daft_dtype.size
        )
    return dtypes.Unknown()  # pragma: no cover


def narwhals_to_native_dtype(  # noqa: PLR0912,C901,PLR0911
    dtype: DType | type[DType], version: Version
) -> daft.DataType:
    dtypes = version.dtypes
    if dtype == dtypes.Float64:
        return DataType.float64()
    if dtype == dtypes.Float32:
        return DataType.float32()
    if dtype in {dtypes.Int128, dtypes.UInt128}:  # pragma: no cover
        msg = "Converting to Int128/UInt128 is not (yet) supported for Daft."
        raise NotImplementedError(msg)
    if dtype == dtypes.Int64:
        return DataType.int64()
    if dtype == dtypes.Int32:
        return DataType.int32()
    if dtype == dtypes.Int16:
        return DataType.int16()
    if dtype == dtypes.Int8:
        return DataType.int8()
    if dtype == dtypes.UInt64:
        return DataType.uint64()
    if dtype == dtypes.UInt32:
        return DataType.uint32()
    if dtype == dtypes.UInt16:
        return DataType.uint16()
    if dtype == dtypes.UInt8:
        return DataType.uint8()
    if dtype == dtypes.String:
        return DataType.string()
    if dtype == dtypes.Boolean:
        return DataType.bool()
    if dtype == dtypes.Object:  # pragma: no cover
        msg = "Converting to Object is not (yet) supported for Daft"
        raise NotImplementedError(msg)
    if dtype == dtypes.Categorical:
        msg = "Converting to Categorical is not (yet) supported for Daft"
        raise NotImplementedError(msg)
    if dtype == dtypes.Enum:
        msg = "Converting to Enum is not (yet) supported for Daft"
        raise NotImplementedError(msg)
    if dtype == dtypes.Date:
        return DataType.date()
    if dtype == dtypes.Time:
        return DataType.time("ns")
    if dtype == dtypes.Binary:
        return DataType.binary()
    if dtype == dtypes.Decimal:  # pragma: no cover
        msg = "Casting to Decimal is not supported yet."
        raise NotImplementedError(msg)
    if isinstance_or_issubclass(dtype, dtypes.Datetime):
        return DataType.timestamp(dtype.time_unit, dtype.time_zone)
    if isinstance_or_issubclass(dtype, dtypes.Duration):
        return DataType.duration(dtype.time_unit)
    if isinstance_or_issubclass(dtype, dtypes.List):
        return DataType.list(narwhals_to_native_dtype(dtype.inner, version))
    if isinstance_or_issubclass(dtype, dtypes.Struct):
        return DataType.struct(
            {
                field.name: narwhals_to_native_dtype(field.dtype, version)
                for field in dtype.fields
            }
        )
    if isinstance_or_issubclass(dtype, dtypes.Array):  # pragma: no cover
        return DataType.fixed_size_list(
            narwhals_to_native_dtype(dtype.inner, version), dtype.size
        )
    msg = f"Unknown dtype: {dtype}"  # pragma: no cover
    raise AssertionError(msg)


def extend_bool(
    value: bool | Iterable[bool],  # noqa: FBT001
    n_match: int,
) -> Sequence[bool]:
    """Ensure the given bool or sequence of bools is the correct length.

    Stolen from https://github.com/pola-rs/polars/blob/b8bfb07a4a37a8d449d6d1841e345817431142df/py-polars/polars/_utils/various.py#L580-L594
    """
    return (value,) * n_match if isinstance(value, bool) else tuple(value)


def evaluate_literal(expr: DaftExpr) -> object:
    return next(expr._metadata.op_nodes_reversed()).kwargs["value"]
